@@ -1,3 +1,5 @@
+# frozen_string_literal: true
+
 Puppet::DataTypes.create_type('Error') do
   interface <<-PUPPET
     type_parameters => {
@@ -13,9 +15,9 @@ Puppet::DataTypes.create_type('Error') do
     functions => {
       message => Callable[[], String[1]]
     }
-    PUPPET
+  PUPPET
 
-  require 'puppet/datatypes/impl/error'
+  require_relative '../../puppet/datatypes/impl/error'
 
   implementation_class Puppet::DataTypes::Error
 end
