@@ -1,13 +1,13 @@
+# frozen_string_literal: true
+
 module Puppet::Pops
 module Types
-
 # A Puppet Language type that wraps sensitive information. The sensitive type is parameterized by
 # the wrapped value type.
 #
 #
 # @api public
-class PSensitiveType < Puppet::Pops::Types::PTypeWithContainedType
-
+class PSensitiveType < PTypeWithContainedType
   class Sensitive
     def initialize(value)
       @value = value
@@ -22,8 +22,22 @@ class PSensitiveType < Puppet::Pops::Types::PTypeWithContainedType
     end
 
     def inspect
-      "#<#{to_s}>"
+      "#<#{self}>"
     end
+
+    def hash
+      @value.hash
+    end
+
+    def ==(other)
+      other.is_a?(Sensitive) &&
+        other.hash == hash
+    end
+    alias eql? ==
+  end
+
+  def self.register_ptype(loader, ir)
+    create_ptype(loader, ir, 'AnyType')
   end
 
   def initialize(type = nil)
@@ -34,9 +48,8 @@ class PSensitiveType < Puppet::Pops::Types::PTypeWithContainedType
     o.is_a?(Sensitive) && @type.instance?(o.unwrap, guard)
   end
 
-  def self.new_function(_, loader)
-    @new_function ||= Puppet::Functions.create_loaded_function(:new_Sensitive, loader) do
-
+  def self.new_function(type)
+    @new_function ||= Puppet::Functions.create_loaded_function(:new_Sensitive, type.loader) do
       dispatch :from_sensitive do
         param 'Sensitive', :value
       end
@@ -59,11 +72,10 @@ class PSensitiveType < Puppet::Pops::Types::PTypeWithContainedType
   private
 
   def _assignable?(o, guard)
-    self.class == o.class && @type.assignable?(o.type, guard)
+    instance_of?(o.class) && @type.assignable?(o.type, guard)
   end
 
   DEFAULT = PSensitiveType.new
 end
-
 end
 end
